"""
Airspeed Velocity benchmark utilities
"""
import sys
import os
import re
import time
import textwrap
import subprocess
import itertools
import random

from asv_runner.benchmarks.mark import SkipNotImplemented


class Benchmark:
    """
    Base class with sensible options
    """


class XPBenchmark(Benchmark):
    """
    Base class for benchmarks that are run on multiple Array API backends
    and devices. Supports multiple devices, jax.jit, and lazy/asynchronous
    evaluation.

    Basic usage
    -----------
    ::
        def myfunc(x):
            return x + 1

        class MyFunc(XPBenchmark):
            def setup(self, backend):
                super().setup(backend, myfunc)
                x = self.xp.arange(5)
                self.x = self.synchronize(x)
                if self.warmup:
                    self.func(self.x)

            def time_myfunc(self, backend):
                self.func(self.x)

    Adding parameters
    -----------------
    In the below example:
    - We add a `size` asv parameter
    - We add a `plus` function parameter which can't be traced by jax.jit

    ::
        def myfunc(x, plus=True):
            return x + 1 if plus else x - 1

        class MyFunc(XPBenchmark):
            param_names = (*XPBenchmark.param_names, "size")
            params = (*XPBenchmark.params, [5, 10])

            def setup(self, backend, size):
                super().setup(backend, myfunc, static_argnames=("plus",))
                x = self.xp.arange(size)
                self.x = self.synchronize(x)
                if self.warmup:
                    self.func(self.x, plus=True)
                    self.func(self.x, plus=False)

            def time_myfunc_plus(self, backend, size):
                self.func(self.x, plus=True)

            def time_myfunc_minus(self, backend, size):
                self.func(self.x, plus=False)
    """
    backends = ["numpy", "array_api_strict", "cupy", "torch:cpu", "torch:cuda",
                "dask.array", "jax.numpy:cpu", "jax.numpy:cuda"]

    # subclasses can override these
    param_names = ("backend",)
    params = (backends, )

    def setup(self, backend, func, *, static_argnums=None, static_argnames=None):
        """Skip benchmark if backend/device combination is not available.
        Configure namespace.
        Potentially wrap func with jax.jit and ensure timings are correct
        for lazy backends.

        Parameters
        ----------
        backend : str
            backend name from asv parameterization
        func : callable
            function to benchmark
        static_argnums : Sequence[int], optional
            Parameter for jax.jit. Note that, unlike in the unit tests,
            we can't use the automatic parameter and return value wrap/unwrap
            from `array_api_extra.testing.lazy_xp_function`, as it comes with a
            substantial performance overhead.
        static_argnames : Sequence[str], optional
            Parameter for jax.jit

        Sets attributes
        ---------------
        backend : str
            As the parameter (for convenience of helper functions)
        xp : namespace
            array namespace, potentially wrapped by array_api_compat
        func : callable
            function to benchmark, potentially wrapped
        warmup : bool
            Whether setup() should run a warmup iteration
        """
        self.backend = backend
        if ":" in backend:
            backend, device = backend.split(":")
        else:
            device = "cuda" if backend == "cupy" else "cpu"

        with safe_import() as array_api_imports:
            # Requires scipy >=1.16
            from scipy._lib._array_api import array_namespace, xp_capabilities_table
            from scipy.conftest import xp_available_backends, xp_known_backends

            if isinstance(xp_available_backends, dict):  # scipy == 1.16
                backends = xp_available_backends
            else:  # scipy >= 1.17
                backends = {p.id: p.values[0] for p in xp_available_backends}
        if array_api_imports.error:
            # On older scipy versions, disregard SCIPY_ARRAY_API
            import numpy as np
            def array_namespace(*args, **kwargs):
                return np
            xp_capabilities_table = {}
            backends = {"numpy": np}
            xp_known_backends = {"numpy"}

        # If new backends are added to conftest.py, you need to add them here too
        assert not xp_known_backends - set(n.split(":")[0] for n in self.backends)

        try:
            xp = backends[backend]
        except KeyError:
            raise SkipNotImplemented(
                f"{backend} not available or skipped by SCIPY_ARRAY_API")

        if func and func in xp_capabilities_table:
            capabilities = xp_capabilities_table[func]
            skips = {n for n, _ in capabilities["skip_backends"]}
            skips |= {n for n, _ in capabilities["xfail_backends"]}
            if (((capabilities["cpu_only"] and device != "cpu")
                 or (capabilities["np_only"] and backend != "numpy"))
                and backend not in capabilities["exceptions"]):
                skips.add(backend)
            if backend in skips:
                raise SkipNotImplemented(f"{backend} skipped by @xp_capabilities")
        else:
            capabilities = {"jax_jit": False}

        # Potentially wrap namespace with array_api_compat
        xp = array_namespace(xp.empty(0))

        self.xp = xp
        self.func = func
        self.warmup = False

        if backend == "torch":
            import torch

            torch.set_default_dtype(torch.float64)
            try:
                torch.empty(0, device=device)
            except (RuntimeError, AssertionError):
                raise SkipNotImplemented(f"{device=} not available")
            torch.set_default_device(device)

            if device == "cuda":
                def wrapper(*args, **kwargs):
                    res = func(*args, **kwargs)
                    torch.cuda.synchronize()
                    return res

                self.func = wrapper

        elif backend == "jax.numpy":
            import jax

            jax.config.update("jax_enable_x64", True)
            try:
                jax_device = jax.devices(device)[0]
            except RuntimeError:
                raise SkipNotImplemented(f"{device=} not available")
            jax.config.update("jax_default_device", jax_device)

            if capabilities["jax_jit"]:
                func = jax.jit(func, static_argnames=static_argnames,
                               static_argnums=static_argnums)
                self.warmup = True

            def wrapper(*args, **kwargs):
                res = func(*args, **kwargs)
                jax.block_until_ready(res)
                return res

            self.func = wrapper

        elif backend == "dask.array":
            import dask

            def wrapper(*args, **kwargs):
                res = func(*args, **kwargs)
                return dask.compute(res)[0]

            self.func = wrapper

        elif backend == "cupy":
            import cupy
            # The default stream is non-blocking.
            # As of CuPy 13.4.1, explicit non-blocking streams
            # are substantially slower.
            # cupy.cuda.Stream(non_blocking=True).use()

            def wrapper(*args, **kwargs):
                res = func(*args, **kwargs)
                cupy.cuda.get_current_stream().synchronize()
                return res

            self.func = wrapper

        else:
            assert backend in ("numpy", "array_api_strict")

    def synchronize(self, *arrays):
        """Wait until the given arrays have finished generating and return a
        synchronized instance of them.
        You need to call this on all arrays that your setup() function creates.
        """
        if self.backend == "dask.array":
            import dask

            arrays = dask.persist(*arrays)
        elif self.backend in ("jax.numpy:cpu", "jax.numpy:cuda"):
            import jax

            jax.block_until_ready(arrays)
        elif self.backend == "torch:cuda":
            import torch

            torch.cuda.synchronize()
        elif self.backend == "cupy":
            import cupy

            cupy.cuda.get_current_stream().synchronize()
        else:
            assert self.backend in ("numpy", "array_api_strict", "torch:cpu")

        return arrays[0] if len(arrays) == 1 else arrays


def is_xslow():
    try:
        return int(os.environ.get('SCIPY_XSLOW', '0'))
    except ValueError:
        return False


class LimitedParamBenchmark(Benchmark):
    """
    Limits parameter combinations to `max_number` choices, chosen
    pseudo-randomly with fixed seed.
    Raises NotImplementedError (skip) if not in active set.
    """
    num_param_combinations = 0

    def setup(self, *args, **kwargs):
        slow = is_xslow()

        if slow:
            # no need to skip
            return

        param_seed = kwargs.pop('param_seed', None)
        if param_seed is None:
            param_seed = 1

        params = kwargs.pop('params', None)
        if params is None:
            params = self.params

        num_param_combinations = kwargs.pop('num_param_combinations', None)
        if num_param_combinations is None:
            num_param_combinations = self.num_param_combinations

        all_choices = list(itertools.product(*params))

        rng = random.Random(param_seed)
        rng.shuffle(all_choices)
        active_choices = all_choices[:num_param_combinations]

        if args not in active_choices:
            raise NotImplementedError("skipped")


def get_max_rss_bytes(rusage):
    """
    Extract the max RSS value in bytes.
    """
    if not rusage:
        return None

    if sys.platform.startswith('linux'):
        # On Linux getrusage() returns ru_maxrss in kilobytes
        # https://man7.org/linux/man-pages/man2/getrusage.2.html
        return rusage.ru_maxrss * 1024
    elif sys.platform == "darwin":
        # on macOS ru_maxrss is in bytes
        return rusage.ru_maxrss
    else:
        # Unknown, just return whatever is here.
        return rusage.ru_maxrss


def run_monitored_wait4(code):
    """
    Run code in a new Python process, and monitor peak memory usage.

    Returns
    -------
    duration : float
        Duration in seconds (including Python startup time)
    peak_memusage : int
        Peak memory usage in bytes of the child Python process

    Notes
    -----
    Works on Unix platforms (Linux, macOS) that have `os.wait4()`.
    """
    code = textwrap.dedent(code)

    start = time.time()
    process = subprocess.Popen([sys.executable, '-c', code])
    pid, returncode, rusage = os.wait4(process.pid, 0)
    duration = time.time() - start
    max_rss_bytes = get_max_rss_bytes(rusage)

    if returncode != 0:
        raise AssertionError(f"Running failed:\n{code}")

    return duration, max_rss_bytes


def run_monitored_proc(code):
    """
    Run code in a new Python process, and monitor peak memory usage.

    Returns
    -------
    duration : float
        Duration in seconds (including Python startup time)
    peak_memusage : float
        Peak memory usage (rough estimate only) in bytes

    """
    if not sys.platform.startswith('linux'):
        raise RuntimeError("Peak memory monitoring only works on Linux")

    code = textwrap.dedent(code)
    process = subprocess.Popen([sys.executable, '-c', code])

    peak_memusage = -1

    start = time.time()
    while True:
        ret = process.poll()
        if ret is not None:
            break

        with open(f'/proc/{process.pid}/status') as f:
            procdata = f.read()

        m = re.search(r'VmRSS:\s*(\d+)\s*kB', procdata, re.S | re.I)
        if m is not None:
            memusage = float(m.group(1)) * 1e3
            peak_memusage = max(memusage, peak_memusage)

        time.sleep(0.01)

    process.wait()

    duration = time.time() - start

    if process.returncode != 0:
        raise AssertionError(f"Running failed:\n{code}")

    return duration, peak_memusage


def run_monitored(code):
    """
    Run code in a new Python process, and monitor peak memory usage.

    Returns
    -------
    duration : float
        Duration in seconds (including Python startup time)
    peak_memusage : float or int
        Peak memory usage (rough estimate only) in bytes

    """

    if hasattr(os, 'wait4'):
        return run_monitored_wait4(code)
    else:
        return run_monitored_proc(code)


def get_mem_info():
    """Get information about available memory"""
    import psutil
    vm = psutil.virtual_memory()
    return {
        "memtotal": vm.total,
        "memavailable": vm.available,
    }


def set_mem_rlimit(max_mem=None):
    """
    Set address space rlimit
    """
    import resource
    if max_mem is None:
        mem_info = get_mem_info()
        max_mem = int(mem_info['memtotal'] * 0.7)
    cur_limit = resource.getrlimit(resource.RLIMIT_AS)
    if cur_limit[0] > 0:
        max_mem = min(max_mem, cur_limit[0])

    try:
        resource.setrlimit(resource.RLIMIT_AS, (max_mem, cur_limit[1]))
    except ValueError:
        # on macOS may raise: current limit exceeds maximum limit
        pass


def with_attributes(**attrs):
    def decorator(func):
        for key, value in attrs.items():
            setattr(func, key, value)
        return func
    return decorator


class safe_import:

    def __enter__(self):
        self.error = False
        return self

    def __exit__(self, type_, value, traceback):
        if type_ is not None:
            self.error = True
            suppress = not (
                os.getenv('SCIPY_ALLOW_BENCH_IMPORT_ERRORS', '1').lower() in
                ('0', 'false') or not issubclass(type_, ImportError))
            return suppress



from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.states import Body
from docutils.statemachine import StringList

from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import nested_parse_with_titles
from sphinx.util.typing import ExtensionMetadata

from tabulate import tabulate

from scipy._lib._array_api_docs_tables import BackendSupportStatus
from scipy._lib._array_api_docs_tables import calculate_table_statistics
from scipy._lib._array_api_docs_tables import make_flat_capabilities_table
from scipy._lib._public_api import PUBLIC_MODULES


def _make_reST_table(
        rows: list[str],
        headers: list[str],
        state: Body,
) -> list[nodes.Node]:
    """Return list of docutils nodes representing a reST table."""
    rst_table = tabulate(rows, headers=headers, tablefmt="rst")
    string_list = StringList(rst_table.splitlines())
    node = nodes.section()
    node.document = state.document
    nested_parse_with_titles(state, string_list, node)
    return node.children


def _get_flat_table_and_backends(
        env: BuildEnvironment, backend_type: str
) -> tuple[list[dict[str, str]], list[str]]:
    if not hasattr(env, "_array_api_capabilities_table_cache"):
        env._array_api_capabilities_table_cache = {}
    cache = env._array_api_capabilities_table_cache
    if backend_type in cache:
        return cache[backend_type]["flat_table"], cache[backend_type]["backends"]

    included_modules = [
        module for module in PUBLIC_MODULES
        # These are extension modules which should never be included. The introspection
        # in make_flat_capabilities_table will fail for these.
        if module not in {"scipy.linalg.cython_blas", "scipy.linalg.cython_lapack"}
    ]

    flat_table = make_flat_capabilities_table(
        included_modules, backend_type
    )

    if not flat_table:
        backends = []
    else:
        backends = [key for key in flat_table[0]
                    if key not in ["module", "function"]]

    cache[backend_type] = {"flat_table": flat_table, "backends": backends}
    return flat_table, backends


class ArrayAPISupportPerModule(SphinxDirective):
    has_content = False
    option_spec = {
        module_name.replace("scipy.", ""): directives.unchanged
        for module_name in PUBLIC_MODULES
    }
    option_spec["backend_type"] = directives.unchanged

    def run(self):
        backend_type = self.options.pop("backend_type")
        flat_table, backends = _get_flat_table_and_backends(self.env, backend_type)
        table_stats = calculate_table_statistics(flat_table)
        modules = self.options.keys()

        rows = []
        headers = ["module"]
        headers += backends

        for module in modules:
            label = self.options.get(module)
            module_text = f":ref:`{module} <{label}>`" if label else module
            info, accurate_count = table_stats[f"scipy.{module}"]
            total = info.pop("total")
            row = [f"{module_text} ({total})"]
            for backend, count in info.items():
                cell_text = f"{count/total:.0%}{'*' if not accurate_count else ''}"
                row.append(cell_text)
            rows.append(row)
        return _make_reST_table(rows, headers, self.state)


class ArrayAPISupportPerFunction(SphinxDirective):
    has_content = False
    option_spec = {
        "module": directives.unchanged, "backend_type": directives.unchanged
    }

    def _get_generated_doc_link_for_function(self, module, func):
        if module == "signal" and func == "czt":
            pagename = "czt-function"
        else:
            pagename = f"scipy.{module}.{func}"
        return (
            f":doc:`{func} <../../../reference/generated/{pagename}>`"
        )

    def run(self):
        backend_type = self.options["backend_type"]
        flat_table, backends = _get_flat_table_and_backends(self.env, backend_type)
        module = self.options["module"]
        relevant_rows = (
            row for row in flat_table if row["module"] == f"scipy.{module}"
        )

        headers = ["function"]
        headers += backends

        new_rows = []
        S = BackendSupportStatus
        missing_xp_capabilities = False
        for row in relevant_rows:
            func = row["function"]
            new_row = [self._get_generated_doc_link_for_function(module, func)]
            for backend in backends:
                supported = row[backend]
                cell_text = ""
                if supported == S.OUT_OF_SCOPE:
                    cell_text = "N/A"
                elif supported == S.YES:
                    cell_text = "✔️"
                elif supported == S.NO:
                    cell_text = "✖"
                else:
                    missing_xp_capabilities = True
                new_row.append(cell_text)
            new_rows.append(new_row)
        table_nodes = _make_reST_table(new_rows, headers, self.state)

        legend = nodes.admonition()
        legend += nodes.title(text="Legend")
        legend += nodes.paragraph(text="✔️ = supported")
        legend += nodes.paragraph(text="✖ = unsupported")
        legend += nodes.paragraph(text="N/A = out-of-scope")
        if missing_xp_capabilities:
            # Only include blank in the legend if there are actually functions
            # missing the xp_capabilities decorator.
            legend += nodes.paragraph(text="blank = not currently documented")
        return [legend] + table_nodes

def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_directive(
        "array-api-support-per-module", ArrayAPISupportPerModule
    )
    app.add_directive(
        "array-api-support-per-function", ArrayAPISupportPerFunction
    )
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Test functions for the sparse.linalg._expm_multiply module."""
import warnings

from functools import partial
from itertools import product

import numpy as np
import pytest
from numpy.testing import (assert_allclose, assert_, assert_equal)
from scipy.sparse import SparseEfficiencyWarning
import scipy.sparse
from scipy.sparse.linalg import aslinearoperator
import scipy.linalg
from scipy.sparse.linalg import expm as sp_expm
from scipy.sparse.linalg._expm_multiply import (_theta, _compute_p_max,
        _onenormest_matrix_power, expm_multiply, _expm_multiply_simple,
        _expm_multiply_interval)
from scipy._lib._util import np_long


IMPRECISE = {np.single, np.csingle}
REAL_DTYPES = (np.intc, np_long, np.longlong,
               np.float32, np.float64, np.longdouble)
COMPLEX_DTYPES = (np.complex64, np.complex128, np.clongdouble)
DTYPES = REAL_DTYPES + COMPLEX_DTYPES


def estimated_warns():
    """If trace is estimated, it should warn.

    We warn that estimation of trace might impact performance.
    All result have to be correct nevertheless!
    """
    return pytest.warns(UserWarning, match="Trace of LinearOperator not available")


def less_than_or_close(a, b):
    return np.allclose(a, b) or (a < b)


class TestExpmActionSimple:
    """
    These tests do not consider the case of multiple time steps in one call.
    """

    def test_theta_monotonicity(self):
        pairs = sorted(_theta.items())
        for (m_a, theta_a), (m_b, theta_b) in zip(pairs[:-1], pairs[1:]):
            assert_(theta_a < theta_b)

    def test_p_max_default(self):
        m_max = 55
        expected_p_max = 8
        observed_p_max = _compute_p_max(m_max)
        assert_equal(observed_p_max, expected_p_max)

    def test_p_max_range(self):
        for m_max in range(1, 55+1):
            p_max = _compute_p_max(m_max)
            assert_(p_max*(p_max - 1) <= m_max + 1)
            p_too_big = p_max + 1
            assert_(p_too_big*(p_too_big - 1) > m_max + 1)

    def test_onenormest_matrix_power(self):
        rng = np.random.RandomState(1234)
        n = 40
        nsamples = 10
        for i in range(nsamples):
            A = scipy.linalg.inv(rng.randn(n, n))
            for p in range(4):
                if not p:
                    M = np.identity(n)
                else:
                    M = np.dot(M, A)
                estimated = _onenormest_matrix_power(A, p)
                exact = np.linalg.norm(M, 1)
                assert_(less_than_or_close(estimated, exact))
                assert_(less_than_or_close(exact, 3*estimated))

    def test_expm_multiply(self):
        rng = np.random.default_rng(1234)
        n = 40
        k = 3
        nsamples = 10
        for i in range(nsamples):
            A = scipy.linalg.inv(rng.standard_normal((n, n)))
            B = rng.standard_normal((n, k))
            observed = expm_multiply(A, B)
            expected = np.dot(sp_expm(A), B)
            assert_allclose(observed, expected)
            with estimated_warns():
                observed = expm_multiply(aslinearoperator(A), B)
            assert_allclose(observed, expected)
            traceA = np.trace(A)
            observed = expm_multiply(aslinearoperator(A), B, traceA=traceA)
            assert_allclose(observed, expected)

    def test_matrix_vector_multiply(self):
        rng = np.random.default_rng(1234)
        n = 40
        nsamples = 10
        for i in range(nsamples):
            A = scipy.linalg.inv(rng.standard_normal((n, n)))
            v = rng.standard_normal(n)
            observed = expm_multiply(A, v)
            expected = np.dot(sp_expm(A), v)
            assert_allclose(observed, expected)
            with estimated_warns():
                observed = expm_multiply(aslinearoperator(A), v)
            assert_allclose(observed, expected)

    def test_scaled_expm_multiply(self):
        rng = np.random.default_rng(1234)
        n = 40
        k = 3
        nsamples = 10
        for i, t in product(range(nsamples), [0.2, 1.0, 1.5]):
            with np.errstate(invalid='ignore'):
                A = scipy.linalg.inv(rng.standard_normal((n, n)))
                B = rng.standard_normal((n, k))
                observed = _expm_multiply_simple(A, B, t=t)
                expected = np.dot(sp_expm(t*A), B)
                assert_allclose(observed, expected)
                with estimated_warns():
                    observed = _expm_multiply_simple(aslinearoperator(A), B, t=t)
                assert_allclose(observed, expected)

    def test_scaled_expm_multiply_single_timepoint(self):
        rng = np.random.default_rng(1234)
        t = 0.1
        n = 5
        k = 2
        A = rng.standard_normal((n, n))
        B = rng.standard_normal((n, k))
        observed = _expm_multiply_simple(A, B, t=t)
        expected = sp_expm(t*A).dot(B)
        assert_allclose(observed, expected)
        with estimated_warns():
            observed = _expm_multiply_simple(aslinearoperator(A), B, t=t)
        assert_allclose(observed, expected)

    def test_sparse_expm_multiply(self):
        rng = np.random.default_rng(1234)
        n = 40
        k = 3
        nsamples = 10
        for i in range(nsamples):
            A = scipy.sparse.random_array((n, n), density=0.05, rng=rng)
            B = rng.standard_normal((n, k))
            observed = expm_multiply(A, B)
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    "splu converted its input to CSC format",
                    SparseEfficiencyWarning,
                )
                warnings.filterwarnings(
                    "ignore",
                    "spsolve is more efficient when sparse b is in the "
                    "CSC matrix format",
                    SparseEfficiencyWarning,
                )
                expected = sp_expm(A).dot(B)
            assert_allclose(observed, expected)
            with estimated_warns():
                observed = expm_multiply(aslinearoperator(A), B)
            assert_allclose(observed, expected)

    def test_complex(self):
        A = np.array([
            [1j, 1j],
            [0, 1j]], dtype=complex)
        B = np.array([1j, 1j])
        observed = expm_multiply(A, B)
        expected = np.array([
            1j * np.exp(1j) + 1j * (1j*np.cos(1) - np.sin(1)),
            1j * np.exp(1j)], dtype=complex)
        assert_allclose(observed, expected)
        with estimated_warns():
            observed = expm_multiply(aslinearoperator(A), B)
        assert_allclose(observed, expected)


class TestExpmActionInterval:

    @pytest.mark.fail_slow(20)
    def test_sparse_expm_multiply_interval(self):
        rng = np.random.default_rng(1234)
        start = 0.1
        stop = 3.2
        n = 40
        k = 3
        endpoint = True
        for num in (14, 13, 2):
            A = scipy.sparse.random_array((n, n), density=0.05, rng=rng)
            B = rng.standard_normal((n, k))
            v = rng.standard_normal((n,))
            for target in (B, v):
                X = expm_multiply(A, target, start=start, stop=stop,
                                  num=num, endpoint=endpoint)
                samples = np.linspace(start=start, stop=stop,
                                      num=num, endpoint=endpoint)
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore",
                        "splu converted its input to CSC format",
                        SparseEfficiencyWarning,
                    )
                    warnings.filterwarnings(
                        "ignore",
                        "spsolve is more efficient when sparse b is in"
                        " the CSC matrix format",
                        SparseEfficiencyWarning,
                    )
                    for solution, t in zip(X, samples):
                        assert_allclose(solution, sp_expm(t*A).dot(target))

    @pytest.mark.fail_slow(20)
    def test_expm_multiply_interval_vector(self):
        rng = np.random.default_rng(1234)
        interval = {'start': 0.1, 'stop': 3.2, 'endpoint': True}
        for num, n in product([14, 13, 2], [1, 2, 5, 20, 40]):
            A = scipy.linalg.inv(rng.standard_normal((n, n)))
            v = rng.standard_normal(n)
            samples = np.linspace(num=num, **interval)
            X = expm_multiply(A, v, num=num, **interval)
            for solution, t in zip(X, samples):
                assert_allclose(solution, sp_expm(t*A).dot(v))
            # test for linear operator with unknown trace -> estimate trace
            with estimated_warns():
                Xguess = expm_multiply(aslinearoperator(A), v, num=num, **interval)
            # test for linear operator with given trace
            Xgiven = expm_multiply(aslinearoperator(A), v, num=num, **interval,
                                   traceA=np.trace(A))
            # test robustness for linear operator with wrong trace
            Xwrong = expm_multiply(aslinearoperator(A), v, num=num, **interval,
                                   traceA=np.trace(A)*5)
            for sol_guess, sol_given, sol_wrong, t in zip(Xguess, Xgiven,
                                                          Xwrong, samples):
                correct = sp_expm(t*A).dot(v)
                assert_allclose(sol_guess, correct)
                assert_allclose(sol_given, correct)
                assert_allclose(sol_wrong, correct)

    @pytest.mark.fail_slow(20)
    def test_expm_multiply_interval_matrix(self):
        rng = np.random.default_rng(1234)
        interval = {'start': 0.1, 'stop': 3.2, 'endpoint': True}
        for num, n, k in product([14, 13, 2], [1, 2, 5, 20, 40], [1, 2]):
            A = scipy.linalg.inv(rng.standard_normal((n, n)))
            B = rng.standard_normal((n, k))
            samples = np.linspace(num=num, **interval)
            X = expm_multiply(A, B, num=num, **interval)
            for solution, t in zip(X, samples):
                assert_allclose(solution, sp_expm(t*A).dot(B))
            with estimated_warns():
                X = expm_multiply(aslinearoperator(A), B, num=num, **interval)
            for solution, t in zip(X, samples):
                assert_allclose(solution, sp_expm(t*A).dot(B))

    def test_sparse_expm_multiply_interval_dtypes(self):
        # Test A & B int
        A = scipy.sparse.diags_array(np.arange(5),format='csr', dtype=int)
        B = np.ones(5, dtype=int)
        Aexpm = scipy.sparse.diags_array(np.exp(np.arange(5)),format='csr')
        BI = np.identity(5, dtype=int)
        BI_sparse = scipy.sparse.csr_array(BI)
        assert_allclose(expm_multiply(A,B,0,1)[-1], Aexpm.dot(B))
        assert_allclose(np.diag(expm_multiply(A, BI_sparse, 0, 1)[-1]), Aexpm.dot(B))

        # Test A complex, B int
        A = scipy.sparse.diags_array(-1j*np.arange(5),format='csr', dtype=complex)
        B = np.ones(5, dtype=int)
        Aexpm = scipy.sparse.diags_array(np.exp(-1j*np.arange(5)),format='csr')
        assert_allclose(expm_multiply(A,B,0,1)[-1], Aexpm.dot(B))
        assert_allclose(np.diag(expm_multiply(A, BI_sparse, 0, 1)[-1]), Aexpm.dot(B))

        # Test A int, B complex
        A = scipy.sparse.diags_array(np.arange(5),format='csr', dtype=int)
        B = np.full(5, 1j, dtype=complex)
        Aexpm = scipy.sparse.diags_array(np.exp(np.arange(5)),format='csr')
        assert_allclose(expm_multiply(A,B,0,1)[-1], Aexpm.dot(B))
        BI = np.identity(5, dtype=complex)*1j
        assert_allclose(
            np.diag(expm_multiply(A, scipy.sparse.csr_array(BI), 0, 1)[-1]),
            Aexpm.dot(B)
        )

    def test_expm_multiply_interval_status_0(self):
        self._help_test_specific_expm_interval_status(0)

    def test_expm_multiply_interval_status_1(self):
        self._help_test_specific_expm_interval_status(1)

    def test_expm_multiply_interval_status_2(self):
        self._help_test_specific_expm_interval_status(2)

    def _help_test_specific_expm_interval_status(self, target_status):
        rng = np.random.RandomState(1234)
        start = 0.1
        stop = 3.2
        num = 13
        endpoint = True
        n = 5
        k = 2
        nrepeats = 10
        nsuccesses = 0
        for num in [14, 13, 2] * nrepeats:
            A = rng.randn(n, n)
            B = rng.randn(n, k)
            status = _expm_multiply_interval(A, B,
                    start=start, stop=stop, num=num, endpoint=endpoint,
                    status_only=True)
            if status == target_status:
                X, status = _expm_multiply_interval(A, B,
                        start=start, stop=stop, num=num, endpoint=endpoint,
                        status_only=False)
                assert_equal(X.shape, (num, n, k))
                samples = np.linspace(start=start, stop=stop,
                        num=num, endpoint=endpoint)
                for solution, t in zip(X, samples):
                    assert_allclose(solution, sp_expm(t*A).dot(B))
                nsuccesses += 1
        if not nsuccesses:
            msg = 'failed to find a status-' + str(target_status) + ' interval'
            raise Exception(msg)


@pytest.mark.parametrize("dtype_a", DTYPES)
@pytest.mark.parametrize("dtype_b", DTYPES)
@pytest.mark.parametrize("b_is_matrix", [False, True])
def test_expm_multiply_dtype(dtype_a, dtype_b, b_is_matrix):
    """Make sure `expm_multiply` handles all numerical dtypes correctly."""
    assert_allclose_ = (partial(assert_allclose, rtol=1.8e-3, atol=1e-5)
                        if {dtype_a, dtype_b} & IMPRECISE else assert_allclose)
    rng = np.random.default_rng(1234)
    # test data
    n = 7
    b_shape = (n, 3) if b_is_matrix else (n, )
    if dtype_a in REAL_DTYPES:
        A = scipy.linalg.inv(rng.random([n, n])).astype(dtype_a)
    else:
        A = scipy.linalg.inv(
            rng.random([n, n]) + 1j*rng.random([n, n])
        ).astype(dtype_a)
    if dtype_b in REAL_DTYPES:
        B = (2*rng.random(b_shape)).astype(dtype_b)
    else:
        B = (rng.random(b_shape) + 1j*rng.random(b_shape)).astype(dtype_b)

    # single application
    sol_mat = expm_multiply(A, B)
    with estimated_warns():
        sol_op = expm_multiply(aslinearoperator(A), B)
    direct_sol = np.dot(sp_expm(A), B)
    assert_allclose_(sol_mat, direct_sol)
    assert_allclose_(sol_op, direct_sol)
    sol_op = expm_multiply(aslinearoperator(A), B, traceA=np.trace(A))
    assert_allclose_(sol_op, direct_sol)

    # for time points
    interval = {'start': 0.1, 'stop': 3.2, 'num': 13, 'endpoint': True}
    samples = np.linspace(**interval)
    X_mat = expm_multiply(A, B, **interval)
    with estimated_warns():
        X_op = expm_multiply(aslinearoperator(A), B, **interval)
    for sol_mat, sol_op, t in zip(X_mat, X_op, samples):
        direct_sol = sp_expm(t*A).dot(B)
        assert_allclose_(sol_mat, direct_sol)
        assert_allclose_(sol_op, direct_sol)



import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as signal

fs, numtaps = 2, 41  # sampling frequency and number of taps
freqs = [0.0, 0.3, 0.6, 1.0]  # corner frequencies
gains = [1.0, 2.0, 0.5, 0.8]  # desired gains at corner frequencies

b = signal.firwin2(numtaps, freqs, gains, fs=fs)  # design filter
f, H = signal.freqz(b, fs=fs)  # calculate frequency response

fg, ax = plt.subplots(1, 1, layout="constrained")  # do the plotting
ax.set(title=f"Frequency Response of {numtaps}-tap Band-pass FIR-Filter",
       ylabel="Gain", xlim=(0, fs/2),
       xlabel=rf"Frequency $f\,$ in hertz (sampling frequency $f_S={fs}\,$Hz)")
ax.plot(freqs, gains, 'ko--', alpha=.5, label="Desired Gains")
ax.plot(f, np.abs(H), label="Response $|H(f)|$")
ax.grid()
ax.legend()
plt.show()



from .common import XPBenchmark, safe_import

with safe_import():
    from scipy._lib.array_api_compat import array_namespace as compat_namespace
    from scipy._lib._array_api import array_namespace


class ArrayNamespace(XPBenchmark):
    def setup(self, backend):
        def f(x):
            _ = array_namespace(x)
            return x
    
        super().setup(backend, f)
        self.x = self.synchronize(self.xp.empty(0))
        # Populate @lru_cache and jax.jit. Note that this benefits all backends.
        self.func(self.x)

    def time_array_namespace(self, backend):
        """scipy wrapper around array_api_compat.array_namespace"""
        array_namespace(self.x)

    def time_compat_namespace(self, backend):
        """Bare array_api_compat.array_namespace"""
        compat_namespace(self.x)

    def time_trivial_func(self, backend):
        """Trivial function that internally calls `xp=array_namespace(*args)`"""
        self.func(self.x)
